@@ -3,27 +3,20 @@ use crate::io::{decode_request_body, decode_request_headers};
 use crate::model::{
     HeaderName, HeaderValue, InvalidHeader, Request, RequestBuilder, Response, Status,
 };
-#[cfg(feature = "rayon")]
-use rayon_core::ThreadPoolBuilder;
+use std::fmt;
 use std::io::{copy, sink, BufReader, BufWriter, Error, ErrorKind, Result, Write};
-use std::net::{TcpListener, TcpStream, ToSocketAddrs};
-use std::sync::Arc;
-#[cfg(feature = "rayon")]
-use std::thread::available_parallelism;
-#[cfg(not(feature = "rayon"))]
-use std::thread::Builder;
+use std::net::{SocketAddr, TcpListener, TcpStream};
+use std::sync::{Arc, Condvar, Mutex};
+use std::thread::{Builder, JoinHandle};
 use std::time::Duration;
 
 /// An HTTP server.
 ///
-/// It currently provides two threading approaches:
-/// - If the `rayon` feature is enabled a thread pool is used.
-///   The number of threads can be set with the [`Server::set_num_threads`] feature.
-///   By default 4 times the number of available logical cores is used.
-/// - If the `rayon` feature is not enabled, a new thread is started on each connection and kept while the client connection is not closed.
-///   Use it at our own risks!
+/// It uses a very simple threading mechanism: a new thread is started on each connection and closed when the client connection is closed.
+/// To avoid crashes it is possible to set an upper bound to the number of concurrent connections using the [`Server::with_max_concurrent_connections`] function.
 ///
 /// ```no_run
+/// use std::net::{Ipv4Addr, Ipv6Addr};
 /// use oxhttp::Server;
 /// use oxhttp::model::{Response, Status};
 /// use std::time::Duration;
@@ -36,19 +29,23 @@ use std::time::Duration;
 ///         Response::builder(Status::NOT_FOUND).build()
 ///     }
 /// });
+/// // We bind the server to localhost on both IPv4 and v6
+/// server = server.bind((Ipv4Addr::LOCALHOST, 8080)).bind((Ipv6Addr::LOCALHOST, 8080));
 /// // Raise a timeout error if the client does not respond after 10s.
-/// server.set_global_timeout(Duration::from_secs(10));
-/// // Listen to localhost:8080
-/// server.listen(("localhost", 8080))?;
+/// server = server.with_global_timeout(Duration::from_secs(10));
+/// // Limits the number of concurrent connections to 128.
+/// server = server.with_max_concurrent_connections(128);
+/// // We spawn the server and block on it
+/// server.spawn()?.join()?;
 /// # Result::<_,Box<dyn std::error::Error>>::Ok(())
 /// ```
 #[allow(missing_copy_implementations)]
 pub struct Server {
     on_request: Arc<dyn Fn(&mut Request) -> Response + Send + Sync + 'static>,
+    socket_addrs: Vec<SocketAddr>,
     timeout: Option<Duration>,
     server: Option<HeaderValue>,
-    #[cfg(feature = "rayon")]
-    num_threads: usize,
+    max_num_thread: Option<usize>,
 }
 
 impl Server {
@@ -57,166 +54,134 @@ impl Server {
     pub fn new(on_request: impl Fn(&mut Request) -> Response + Send + Sync + 'static) -> Self {
         Self {
             on_request: Arc::new(on_request),
+            socket_addrs: Vec::new(),
             timeout: None,
             server: None,
-            #[cfg(feature = "rayon")]
-            num_threads: available_parallelism().map_or(4, |c| c.get()) * 4,
+            max_num_thread: None,
         }
     }
 
+    /// Ask the server to listen to a given socket when spawned.
+    pub fn bind(mut self, addr: impl Into<SocketAddr>) -> Self {
+        let addr = addr.into();
+        if !self.socket_addrs.contains(&addr) {
+            self.socket_addrs.push(addr);
+        }
+        self
+    }
+
     /// Sets the default value for the [`Server`](https://httpwg.org/http-core/draft-ietf-httpbis-semantics-latest.html#field.server) header.
     #[inline]
-    pub fn set_server_name(
-        &mut self,
+    pub fn with_server_name(
+        mut self,
         server: impl Into<String>,
-    ) -> std::result::Result<(), InvalidHeader> {
+    ) -> std::result::Result<Self, InvalidHeader> {
         self.server = Some(HeaderValue::try_from(server.into())?);
-        Ok(())
+        Ok(self)
     }
 
     /// Sets the global timeout value (applies to both read and write).
     #[inline]
-    pub fn set_global_timeout(&mut self, timeout: Duration) {
+    pub fn with_global_timeout(mut self, timeout: Duration) -> Self {
         self.timeout = Some(timeout);
+        self
     }
 
-    /// Sets that the server should use a thread pool with this number of threads.
-    #[cfg(feature = "rayon")]
+    /// Sets the number maximum number of threads this server can spawn.
     #[inline]
-    pub fn set_num_threads(&mut self, num_threads: usize) {
-        self.num_threads = num_threads;
+    pub fn with_max_concurrent_connections(mut self, max_num_thread: usize) -> Self {
+        self.max_num_thread = Some(max_num_thread);
+        self
     }
 
-    /// Runs the server by listening to `address`.
-    #[cfg(feature = "rayon")]
-    pub fn listen(&self, address: impl ToSocketAddrs) -> Result<()> {
-        let thread_pool = ThreadPoolBuilder::new()
-            .num_threads(self.num_threads)
-            .thread_name(|i| format!("OxHTTP server thread {i}"))
-            .panic_handler(|error| eprintln!("Panic on OxHTTP server thread: {error:?}"))
-            .build()
-            .map_err(|e| Error::new(ErrorKind::Other, e))?;
-        let listeners = open_tcp(address)?;
-        if self.num_threads < listeners.len() + 1 {
-            return Err(Error::new(
-                ErrorKind::Other,
-                format!("The thread pool only provides {} threads whereas the server is listening to {} connections. There are not enough threads for the server to work properly.", self.num_threads, listeners.len()),
-            ));
-        }
-        thread_pool.scope(|p| {
-            for listener in listeners {
-                p.spawn(move |p| {
-                    for stream in listener.incoming() {
-                        match stream {
-                            Ok(stream) => {
-                                let on_request = self.on_request.clone();
-                                let timeout = self.timeout;
-                                let server = self.server.clone();
-                                p.spawn(move |_| {
-                                    let peer = match stream.peer_addr() {
-                                        Ok(peer) => peer,
-                                        Err(error) => {
-                                            eprintln!("OxHTTP TCP error when attempting to get the peer address: {error}");
-                                            return;
-                                        }
-                                    };
-                                    if let Err(error) = accept_request(stream, on_request, timeout, server)
-                                    {
-                                        eprintln!(
-                                            "OxHTTP TCP error when writing response to {peer}: {error}"
-                                        );
-                                    }
-                                })
-                            },
-                            Err(error) => {
-                                eprintln!("OxHTTP TCP error when opening stream: {error}");
-                            }
-                        }
-                    }
-                })
-            }
-        });
-        Ok(())
-    }
-
-    /// Runs the server by listening to `address`.
-    #[cfg(not(feature = "rayon"))]
-    pub fn listen(&self, address: impl ToSocketAddrs) -> Result<()> {
-        // TODO: use scope?
+    /// Spawns the server by listening to the given addresses.
+    ///
+    /// Note that this is not blocking.
+    /// To wait for the server to terminate indefinitely, call [`join`](ListeningServer::join) on the result.
+    pub fn spawn(self) -> Result<ListeningServer> {
         let timeout = self.timeout;
-        let threads = open_tcp(address)?
-            .into_iter()
-            .map(|listener| {
-                let on_request = self.on_request.clone();
-                let server = self.server.clone();
-                Builder::new().name(format!("Thread listening to {}", listener.local_addr()?)).spawn(move || {
-                    for stream in listener.incoming() {
-                        match stream {
-                            Ok(stream) => {
-                                let on_request = on_request.clone();
-                                let server = server.clone();
-                                if let Err(error) =  Builder::new().spawn(move || {
-                                    let peer = match stream.peer_addr() {
+        let thread_limit = self.max_num_thread.map(Semaphore::new);
+        let listener_threads = self.socket_addrs
+                .into_iter()
+                .map(|listener_addr| {
+                    let listener = TcpListener::bind(listener_addr)?;
+                    let thread_name = format!("{}: listener thread of OxHTTP", listener_addr);
+                    let thread_limit = thread_limit.clone();
+                    let on_request = Arc::clone(&self.on_request);
+                    let server = self.server.clone();
+                    Builder::new().name(thread_name).spawn(move || {
+                        for stream in listener.incoming() {
+                            match stream {
+                                Ok(stream) => {
+                                    let peer_addr = match stream.peer_addr() {
                                         Ok(peer) => peer,
                                         Err(error) => {
                                             eprintln!("OxHTTP TCP error when attempting to get the peer address: {error}");
-                                            return;
+                                            continue;
                                         }
                                     };
-                                    if let Err(error) =
-                                        accept_request(stream, on_request, timeout, server)
-                                    {
-                                        eprintln!(
-                                            "OxHTTP TCP error when writing response to {peer}: {error}"
-                                        )
+                                    let thread_name = format!("{}: responding thread of OxHTTP", peer_addr);
+                                    let thread_guard = thread_limit.as_ref().map(|s| s.lock());
+                                    let on_request = Arc::clone(&on_request);
+                                    let server = server.clone();
+                                    if let Err(error) = Builder::new().name(thread_name).spawn(
+                                        move || {
+                                            if let Err(error) =
+                                                accept_request(stream, &*on_request, timeout, &server)
+                                            {
+                                                eprintln!(
+                                                    "OxHTTP TCP error when writing response to {peer_addr}: {error}"
+                                                )
+                                            }
+                                            drop(thread_guard);
+                                        }
+                                    ) {
+                                        eprintln!("OxHTTP thread spawn error: {error}");
                                     }
-                                }) {
-                                    eprintln!("OxHTTP thread spawn error: {error}");
+                                }
+                                Err(error) => {
+                                    eprintln!("OxHTTP TCP error when opening stream: {error}");
                                 }
                             }
-                            Err(error) => {
-                                eprintln!("OxHTTP TCP error when opening stream: {error}");
-                            }
                         }
-                    }
+                    })
                 })
-            })
-            .collect::<Result<Vec<_>>>()?;
-        for thread in threads {
-            thread
-                .join()
-                .map_err(|_| Error::new(ErrorKind::Other, "The server thread panicked"))?;
-        }
-        Ok(())
+                .collect::<Result<Vec<_>>>()?;
+        Ok(ListeningServer {
+            threads: listener_threads,
+        })
     }
 }
 
-fn open_tcp(address: impl ToSocketAddrs) -> Result<Vec<TcpListener>> {
-    let mut listeners = Vec::new();
-    let mut last_error = None;
-    for address in address.to_socket_addrs()? {
-        match TcpListener::bind(address) {
-            Ok(listener) => listeners.push(listener),
-            Err(e) => last_error = Some(e),
+/// Handle to a running server created by [`Server::spawn`].
+pub struct ListeningServer {
+    threads: Vec<JoinHandle<()>>,
+}
+
+impl ListeningServer {
+    /// Join the server threads and wait for them indefinitely except in case of crash.
+    pub fn join(self) -> Result<()> {
+        for thread in self.threads {
+            thread.join().map_err(|e| {
+                Error::new(
+                    ErrorKind::Other,
+                    if let Ok(e) = e.downcast::<&dyn fmt::Display>() {
+                        format!("The server thread panicked with error: {e}")
+                    } else {
+                        "The server thread panicked with an unknown error".into()
+                    },
+                )
+            })?;
         }
-    }
-    if listeners.is_empty() {
-        Err(last_error.unwrap_or_else(|| {
-            Error::new(
-                ErrorKind::InvalidInput,
-                "could not resolve to any addresses",
-            )
-        }))
-    } else {
-        Ok(listeners)
+        Ok(())
     }
 }
 
 fn accept_request(
     mut stream: TcpStream,
-    on_request: Arc<dyn Fn(&mut Request) -> Response>,
+    on_request: &dyn Fn(&mut Request) -> Response,
     timeout: Option<Duration>,
-    server: Option<HeaderValue>,
+    server: &Option<HeaderValue>,
 ) -> Result<()> {
     stream.set_read_timeout(timeout)?;
     stream.set_write_timeout(timeout)?;
@@ -230,7 +195,7 @@ fn accept_request(
                 if let Some(expect) = request.header(&HeaderName::EXPECT).cloned() {
                     if expect.eq_ignore_ascii_case(b"100-continue") {
                         stream.write_all(b"HTTP/1.1 100 Continue\r\n\r\n")?;
-                        read_body_and_build_response(request, reader, on_request.as_ref())
+                        read_body_and_build_response(request, reader, on_request)
                     } else {
                         (
                             build_text_response(
@@ -244,7 +209,7 @@ fn accept_request(
                         )
                     }
                 } else {
-                    read_body_and_build_response(request, reader, on_request.as_ref())
+                    read_body_and_build_response(request, reader, on_request)
                 }
             }
             Err(error) => {
@@ -258,9 +223,17 @@ fn accept_request(
         connection_state = new_connection_state;
 
         // Additional headers
-        set_header_fallback(&mut response, HeaderName::SERVER, &server);
+        if let Some(server) = server {
+            if !response.headers().contains(&HeaderName::SERVER) {
+                response
+                    .headers_mut()
+                    .set(HeaderName::SERVER, server.clone())
+            }
+        }
 
-        encode_response(&mut response, BufWriter::new(&mut stream))?;
+        stream = encode_response(&mut response, BufWriter::new(stream))?
+            .into_inner()
+            .map_err(|e| e.into_error())?;
     }
     Ok(())
 }
@@ -287,7 +260,7 @@ fn read_body_and_build_response(
                     .header(&HeaderName::CONNECTION)
                     .and_then(|v| {
                         v.eq_ignore_ascii_case(b"close")
-                            .then(|| ConnectionState::Close)
+                            .then_some(ConnectionState::Close)
                     })
                     .unwrap_or(ConnectionState::KeepAlive);
                 (response, connection_state)
@@ -315,18 +288,51 @@ fn build_text_response(status: Status, text: String) -> Response {
         .with_body(text)
 }
 
-fn set_header_fallback(
-    response: &mut Response,
-    header_name: HeaderName,
-    header_value: &Option<HeaderValue>,
-) {
-    if let Some(header_value) = header_value {
-        if !response.headers().contains(&header_name) {
-            response
-                .headers_mut()
-                .set(header_name, header_value.clone())
+/// Dumb semaphore allowing to overflow capacity
+#[derive(Clone)]
+struct Semaphore {
+    inner: Arc<InnerSemaphore>,
+}
+
+struct InnerSemaphore {
+    count: Mutex<usize>,
+    capacity: usize,
+    condvar: Condvar,
+}
+
+impl Semaphore {
+    fn new(capacity: usize) -> Self {
+        Self {
+            inner: Arc::new(InnerSemaphore {
+                count: Mutex::new(0),
+                capacity,
+                condvar: Condvar::new(),
+            }),
         }
     }
+
+    fn lock(&self) -> SemaphoreGuard {
+        let data = &self.inner;
+        *data
+            .condvar
+            .wait_while(data.count.lock().unwrap(), |count| *count >= data.capacity)
+            .unwrap() += 1;
+        SemaphoreGuard {
+            inner: Arc::clone(&self.inner),
+        }
+    }
+}
+
+struct SemaphoreGuard {
+    inner: Arc<InnerSemaphore>,
+}
+
+impl Drop for SemaphoreGuard {
+    fn drop(&mut self) {
+        let data = &self.inner;
+        *data.count.lock().unwrap() -= 1;
+        data.condvar.notify_one();
+    }
 }
 
 #[cfg(test)]
@@ -334,7 +340,8 @@ mod tests {
     use super::*;
     use crate::model::Status;
     use std::io::Read;
-    use std::thread::{sleep, spawn};
+    use std::net::{Ipv4Addr, Ipv6Addr};
+    use std::thread::sleep;
 
     #[test]
     fn test_regular_http_operations() -> Result<()> {
@@ -371,18 +378,19 @@ mod tests {
         requests: impl IntoIterator<Item = &'static str>,
         responses: impl IntoIterator<Item = &'static str>,
     ) -> Result<()> {
-        spawn(move || {
-            let mut server = Server::new(|request| {
-                if request.url().path() == "/" {
-                    Response::builder(Status::OK).with_body("home")
-                } else {
-                    Response::builder(Status::NOT_FOUND).build()
-                }
-            });
-            server.set_server_name("OxHTTP/1.0").unwrap();
-            server.set_global_timeout(Duration::from_secs(1));
-            server.listen(("localhost", server_port)).unwrap();
-        });
+        Server::new(|request| {
+            if request.url().path() == "/" {
+                Response::builder(Status::OK).with_body("home")
+            } else {
+                Response::builder(Status::NOT_FOUND).build()
+            }
+        })
+        .bind((Ipv4Addr::LOCALHOST, server_port))
+        .bind((Ipv6Addr::LOCALHOST, server_port))
+        .with_server_name("OxHTTP/1.0")
+        .unwrap()
+        .with_global_timeout(Duration::from_secs(1))
+        .spawn()?;
         sleep(Duration::from_millis(100)); // Makes sure the server is up
         let mut stream = TcpStream::connect((request_host, server_port))?;
         for (request, response) in requests.into_iter().zip(responses) {
@@ -393,4 +401,33 @@ mod tests {
         }
         Ok(())
     }
+
+    #[test]
+    fn test_thread_limit() -> Result<()> {
+        let server_port = 9996;
+        let request = b"GET / HTTP/1.1\nhost: localhost:9999\n\n";
+        let response = b"HTTP/1.1 200 OK\r\nserver: OxHTTP/1.0\r\ncontent-length: 4\r\n\r\nhome";
+        Server::new(|_| Response::builder(Status::OK).with_body("home"))
+            .bind((Ipv4Addr::LOCALHOST, server_port))
+            .bind((Ipv6Addr::LOCALHOST, server_port))
+            .with_server_name("OxHTTP/1.0")
+            .unwrap()
+            .with_global_timeout(Duration::from_secs(1))
+            .with_max_concurrent_connections(2)
+            .spawn()?;
+        sleep(Duration::from_millis(100)); // Makes sure the server is up
+        let streams = (0..128)
+            .map(|_| {
+                let mut stream = TcpStream::connect(("localhost", server_port))?;
+                stream.write_all(request)?;
+                Ok(stream)
+            })
+            .collect::<Result<Vec<_>>>()?;
+        for mut stream in streams {
+            let mut output = vec![b'\0'; response.len()];
+            stream.read_exact(&mut output)?;
+            assert_eq!(output, response);
+        }
+        Ok(())
+    }
 }
